@@ -3,7 +3,7 @@ use std::{
     fs::File,
     io,
     ops::{Add, RangeBounds, Sub},
-    path::{Path, PathBuf},
+    path::{Component, Path, PathBuf},
     process::{Command, Stdio},
     str::CharIndices,
 };
@@ -86,6 +86,7 @@ pub fn find_path_and_position_at(text: &str, index: usize) -> (&str, Option<Buff
     }
 }
 
+// TODO: parse unicode database and implement this
 pub fn char_display_len(_: char) -> u8 {
     1
 }
@@ -561,12 +562,13 @@ impl BufferContent {
             self.line_display_lens.push(DisplayLen::zero());
         }
 
+        let byte_order_mark = b"\xef\xbb\xbf";
         if self.lines[0]
             .as_str()
             .as_bytes()
-            .starts_with(b"\xef\xbb\xbf")
+            .starts_with(byte_order_mark)
         {
-            self.lines[0].delete_range(&mut self.line_display_lens[0], ..3);
+            self.lines[0].delete_range(&mut self.line_display_lens[0], ..byte_order_mark.len());
         }
 
         Ok(())
@@ -924,15 +926,21 @@ impl From<io::Error> for BufferReadError {
     }
 }
 
-pub struct BufferWriteError;
+pub enum BufferWriteError {
+    SavingDisabled,
+    CouldNotWriteToFile,
+}
 impl fmt::Display for BufferWriteError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.write_str("could not write to file")
+        match self {
+            Self::SavingDisabled => f.write_str("buffer has saving disabled"),
+            Self::CouldNotWriteToFile => f.write_str("could not write to file"),
+        }
     }
 }
 impl From<io::Error> for BufferWriteError {
     fn from(_: io::Error) -> Self {
-        Self
+        Self::CouldNotWriteToFile
     }
 }
 
@@ -1022,6 +1030,16 @@ impl Buffer {
         self.handle
     }
 
+    pub fn set_path(&mut self, path: &Path) {
+        self.path.clear();
+        let mut components = path.components();
+        match components.next() {
+            Some(Component::CurDir) => self.path.push(components.as_path()),
+            Some(_) => self.path.push(path),
+            None => (),
+        }
+    }
+
     pub fn highlighted(&self) -> &HighlightedBuffer {
         &self.highlighted
     }
@@ -1332,9 +1350,7 @@ impl Buffer {
             handle: self.handle,
         });
 
-        if !self.path.starts_with(help::HELP_PREFIX)
-            && (!self.properties.saving_enabled || !self.properties.is_file)
-        {
+        if !self.path.starts_with(help::HELP_PREFIX) && !self.properties.is_file {
             return Ok(());
         }
 
@@ -1377,15 +1393,14 @@ impl Buffer {
             Some(path) => {
                 self.properties.saving_enabled = true;
                 self.properties.is_file = true;
-                self.path.clear();
-                self.path.push(path);
+                self.set_path(path);
                 true
             }
             None => false,
         };
 
         if !self.properties.saving_enabled {
-            return Ok(());
+            return Err(BufferWriteError::SavingDisabled);
         }
 
         if self.properties.is_file {
@@ -1444,6 +1459,19 @@ impl BufferCollection {
         buffer
     }
 
+    pub fn try_get(&self, handle: BufferHandle) -> Option<&Buffer> {
+        let index = handle.0 as usize;
+        if self.buffers.len() <= index {
+            return None;
+        }
+        let buffer = &self.buffers[index];
+        if buffer.alive {
+            Some(buffer)
+        } else {
+            None
+        }
+    }
+
     pub fn get(&self, handle: BufferHandle) -> &Buffer {
         &self.buffers[handle.0 as usize]
     }
@@ -1453,9 +1481,11 @@ impl BufferCollection {
     }
 
     pub fn find_with_path(&self, buffers_root: &Path, path: &Path) -> Option<BufferHandle> {
-        if path.as_os_str().is_empty() {
-            return None;
-        }
+        let mut components = path.components();
+        let path = match components.next()? {
+            Component::CurDir => components.as_path(),
+            _ => path,
+        };
 
         for buffer in self.iter() {
             let buffer_path = buffer.path.as_path();
@@ -1510,6 +1540,30 @@ impl BufferCollection {
         }
     }
 
+    pub(crate) fn on_buffer_insert_text(
+        &mut self,
+        buffer_handle: BufferHandle,
+        range: BufferRange,
+    ) {
+        for process in self.insert_processes.iter_mut() {
+            if process.alive && process.buffer_handle == buffer_handle {
+                process.position = process.position.insert(range);
+            }
+        }
+    }
+
+    pub(crate) fn on_buffer_delete_text(
+        &mut self,
+        buffer_handle: BufferHandle,
+        range: BufferRange,
+    ) {
+        for process in self.insert_processes.iter_mut() {
+            if process.alive && process.buffer_handle == buffer_handle {
+                process.position = process.position.delete(range);
+            }
+        }
+    }
+
     pub fn spawn_insert_process(
         &mut self,
         platform: &mut Platform,
@@ -1600,17 +1654,10 @@ impl BufferCollection {
         let texts = process.output_residual_bytes.receive_bytes(&mut buf, bytes);
 
         let buffer = &mut self.buffers[process.buffer_handle.0 as usize];
-        process.position = buffer.content().saturate_position(process.position);
         let mut position = process.position;
-
         for text in texts {
             let insert_range = buffer.insert_text(word_database, position, text, events);
             position = position.insert(insert_range);
-            for process in &mut self.insert_processes {
-                if process.buffer_handle == buffer.handle() {
-                    process.position = process.position.insert(insert_range);
-                }
-            }
         }
     }
 
@@ -1645,26 +1692,48 @@ mod tests {
 
     #[test]
     fn test_find_path_at() {
+        let text = "file.ext:12";
+        assert_eq!(
+            ("file.ext", Some(BufferPosition::line_col(11, 0))),
+            find_path_and_position_at(text, 0),
+        );
+        assert_eq!(
+            ("file.ext", Some(BufferPosition::line_col(11, 0))),
+            find_path_and_position_at(text, 3),
+        );
+        assert_eq!(
+            ("file.ext", Some(BufferPosition::line_col(11, 0))),
+            find_path_and_position_at(text, 4),
+        );
+        assert_eq!(
+            ("file.ext", Some(BufferPosition::line_col(11, 0))),
+            find_path_and_position_at(text, 5),
+        );
+        assert_eq!(
+            ("file.ext", Some(BufferPosition::line_col(11, 0))),
+            find_path_and_position_at(text, 8),
+        );
+
         let text = "/path/file:45";
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(44, 0))),
-            find_path_and_position_at(text, 0)
+            find_path_and_position_at(text, 0),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(44, 0))),
-            find_path_and_position_at(text, 1)
+            find_path_and_position_at(text, 1),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(44, 0))),
-            find_path_and_position_at(text, text.len())
+            find_path_and_position_at(text, text.len()),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(44, 0))),
-            find_path_and_position_at(text, 3)
+            find_path_and_position_at(text, 3),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(44, 0))),
-            find_path_and_position_at(text, 8)
+            find_path_and_position_at(text, 8),
         );
 
         let text = "xx /path/file:";
@@ -1674,41 +1743,41 @@ mod tests {
         assert_eq!(("/path/file", None), find_path_and_position_at(text, 3));
         assert_eq!(
             ("/path/file", None),
-            find_path_and_position_at(text, text.len() - 1)
+            find_path_and_position_at(text, text.len() - 1),
         );
         assert_eq!(
             ("/path/file", None),
-            find_path_and_position_at(text, text.len())
+            find_path_and_position_at(text, text.len()),
         );
 
         let text = "xx /path/file:3xx";
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(2, 0))),
-            find_path_and_position_at(text, 3)
+            find_path_and_position_at(text, 3),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(2, 0))),
-            find_path_and_position_at(text, text.len() - 5)
+            find_path_and_position_at(text, text.len() - 5),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(2, 0))),
-            find_path_and_position_at(text, text.len() - 4)
+            find_path_and_position_at(text, text.len() - 4),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(2, 0))),
-            find_path_and_position_at(text, text.len() - 3)
+            find_path_and_position_at(text, text.len() - 3),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(2, 0))),
-            find_path_and_position_at(text, text.len() - 2)
+            find_path_and_position_at(text, text.len() - 2),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(2, 0))),
-            find_path_and_position_at(text, text.len() - 1)
+            find_path_and_position_at(text, text.len() - 1),
         );
         assert_eq!(
             ("/path/file", Some(BufferPosition::line_col(2, 0))),
-            find_path_and_position_at(text, text.len())
+            find_path_and_position_at(text, text.len()),
         );
 
         let text = "c:/absolute/path/file";
@@ -1717,61 +1786,61 @@ mod tests {
         assert_eq!((text, None), find_path_and_position_at(text, 2));
         assert_eq!(
             (text, None),
-            find_path_and_position_at(text, text.len() - 1)
+            find_path_and_position_at(text, text.len() - 1),
         );
         assert_eq!(
             (text, None),
-            find_path_and_position_at(text, text.len() - 2)
+            find_path_and_position_at(text, text.len() - 2),
         );
 
         let text = "c:/absolute/path/file:4";
         let path = "c:/absolute/path/file";
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 0))),
-            find_path_and_position_at(text, 0)
+            find_path_and_position_at(text, 0),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 0))),
-            find_path_and_position_at(text, 1)
+            find_path_and_position_at(text, 1),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 0))),
-            find_path_and_position_at(text, 2)
+            find_path_and_position_at(text, 2),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 0))),
-            find_path_and_position_at(text, 3)
+            find_path_and_position_at(text, 3),
         );
 
         let text = "xx c:/absolute/path/file:4,5xx";
         let path = "c:/absolute/path/file";
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 4))),
-            find_path_and_position_at(text, 3)
+            find_path_and_position_at(text, 3),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 4))),
-            find_path_and_position_at(text, 4)
+            find_path_and_position_at(text, 4),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 4))),
-            find_path_and_position_at(text, 5)
+            find_path_and_position_at(text, 5),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 4))),
-            find_path_and_position_at(text, 24)
+            find_path_and_position_at(text, 24),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 4))),
-            find_path_and_position_at(text, 25)
+            find_path_and_position_at(text, 25),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 4))),
-            find_path_and_position_at(text, 26)
+            find_path_and_position_at(text, 26),
         );
         assert_eq!(
             (path, Some(BufferPosition::line_col(3, 4))),
-            find_path_and_position_at(text, 27)
+            find_path_and_position_at(text, 27),
         );
     }
 
