@@ -2,11 +2,12 @@ use std::{fmt, process::Command};
 
 use crate::{
     buffer::char_display_len,
-    command::{CommandManager, CommandTokenizer},
-    editor::{BufferedKeys, EditorContext, EditorFlow, KeysIterator},
+    command::CommandTokenizer,
+    editor::{BufferedKeys, KeysIterator},
     events::{KeyParseAllError, KeyParser},
     mode::ModeKind,
-    platform::{Key, Platform},
+    picker::Picker,
+    platform::{Key, KeyCode, Platform},
     word_database::{WordIter, WordKind},
 };
 
@@ -144,13 +145,51 @@ impl ReadLine {
         keys_iter: &mut KeysIterator,
     ) -> ReadLinePoll {
         match keys_iter.next(buffered_keys) {
-            Key::Esc | Key::Ctrl('c') => ReadLinePoll::Canceled,
-            Key::Enter | Key::Ctrl('m') => ReadLinePoll::Submitted,
-            Key::Home | Key::Ctrl('u') => {
+            Key {
+                code: KeyCode::Esc,
+                control: false,
+                alt: false,
+                ..
+            }
+            | Key {
+                code: KeyCode::Char('c'),
+                shift: false,
+                control: true,
+                alt: false,
+            } => ReadLinePoll::Canceled,
+            Key {
+                code: KeyCode::Char('\n'),
+                control: false,
+                alt: false,
+                ..
+            }
+            | Key {
+                code: KeyCode::Char('m'),
+                shift: false,
+                control: true,
+                alt: false,
+            } => ReadLinePoll::Submitted,
+            Key {
+                code: KeyCode::Home,
+                shift: false,
+                control: false,
+                alt: false,
+            }
+            | Key {
+                code: KeyCode::Char('u'),
+                shift: false,
+                control: true,
+                alt: false,
+            } => {
                 self.input.clear();
                 ReadLinePoll::Pending
             }
-            Key::Ctrl('w') => {
+            Key {
+                code: KeyCode::Char('w'),
+                shift: false,
+                control: true,
+                alt: false,
+            } => {
                 let mut words = WordIter(&self.input);
                 (&mut words)
                     .filter(|w| w.kind == WordKind::Identifier)
@@ -159,20 +198,41 @@ impl ReadLine {
                 self.input.truncate(len);
                 ReadLinePoll::Pending
             }
-            Key::Backspace | Key::Ctrl('h') => {
+            Key {
+                code: KeyCode::Backspace,
+                shift: false,
+                control: false,
+                alt: false,
+            }
+            | Key {
+                code: KeyCode::Char('h'),
+                shift: false,
+                control: true,
+                alt: false,
+            } => {
                 if let Some((last_char_index, _)) = self.input.char_indices().next_back() {
                     self.input.truncate(last_char_index);
                 }
                 ReadLinePoll::Pending
             }
-            Key::Ctrl('y') => {
+            Key {
+                code: KeyCode::Char('y'),
+                shift: false,
+                control: true,
+                alt: false,
+            } => {
                 let mut text = string_pool.acquire();
                 platform.read_from_clipboard(&mut text);
                 self.input.push_str(&text);
                 string_pool.release(text);
                 ReadLinePoll::Pending
             }
-            Key::Char(c) => {
+            Key {
+                code: KeyCode::Char(c),
+                control: false,
+                alt: false,
+                ..
+            } => {
                 self.input.push(c);
                 ReadLinePoll::Pending
             }
@@ -360,6 +420,15 @@ impl RegisterKey {
         }
     }
 
+    pub fn from_str(key: &str) -> Option<Self> {
+        let mut chars = key.chars();
+        let c = chars.next()?;
+        if chars.next().is_some() {
+            return None;
+        }
+        Self::from_char(c)
+    }
+
     pub fn as_u8(&self) -> u8 {
         self.0 + b'a'
     }
@@ -388,6 +457,74 @@ impl RegisterCollection {
     }
 }
 
+#[derive(Default)]
+pub(crate) struct PickerEntriesProcessBuf {
+    buf: Vec<u8>,
+    waiting_for_process: bool,
+}
+impl PickerEntriesProcessBuf {
+    pub(crate) fn on_process_spawned(&mut self) {
+        self.waiting_for_process = true;
+    }
+
+    pub(crate) fn on_process_output(
+        &mut self,
+        picker: &mut Picker,
+        read_line: &ReadLine,
+        bytes: &[u8],
+    ) {
+        if !self.waiting_for_process {
+            return;
+        }
+
+        self.buf.extend_from_slice(bytes);
+
+        {
+            let mut entry_adder = picker.add_custom_filtered_entries(read_line.input());
+            if let Some(i) = self.buf.iter().rposition(|&b| b == b'\n') {
+                for line in self
+                    .buf
+                    .drain(..i + 1)
+                    .as_slice()
+                    .split(|&b| matches!(b, b'\n' | b'\r'))
+                {
+                    if line.is_empty() {
+                        continue;
+                    }
+                    if let Ok(line) = std::str::from_utf8(line) {
+                        entry_adder.add(line);
+                    }
+                }
+            }
+        }
+
+        picker.move_cursor(0);
+    }
+
+    pub(crate) fn on_process_exit(&mut self, picker: &mut Picker, read_line: &ReadLine) {
+        if !self.waiting_for_process {
+            return;
+        }
+
+        self.waiting_for_process = false;
+
+        {
+            let mut entry_adder = picker.add_custom_filtered_entries(read_line.input());
+            for line in self.buf.split(|&b| b == b'\n') {
+                if line.is_empty() {
+                    continue;
+                }
+                if let Ok(line) = std::str::from_utf8(line) {
+                    entry_adder.add(line);
+                }
+            }
+        }
+
+        self.buf.clear();
+        picker.move_cursor(0);
+    }
+}
+
 // FNV-1a : https://en.wikipedia.org/wiki/Fowler–Noll–Vo_hash_function
 pub const fn hash_bytes(mut bytes: &[u8]) -> u64 {
     let mut hash: u64 = 0xcbf29ce484222325;
@@ -467,49 +604,15 @@ impl ResidualStrBytes {
 }
 
 pub fn parse_process_command(command: &str) -> Option<Command> {
-    let mut tokenizer = CommandTokenizer(command);
-    let name = tokenizer.next()?;
+    let mut tokens = CommandTokenizer(command);
+    let name = tokens.next()?.slice;
     let mut command = Command::new(name);
-    for arg in tokenizer {
-        command.arg(arg);
+    for arg in tokens {
+        command.arg(arg.slice);
     }
     Some(command)
 }
 
-pub fn load_config(ctx: &mut EditorContext, config_name: &str, config_content: &str) -> EditorFlow {
-    for (line_index, line) in config_content.lines().enumerate() {
-        if line.is_empty() || line.starts_with('#') {
-            continue;
-        }
-
-        let mut command = ctx.editor.string_pool.acquire_with(line);
-        let result = CommandManager::try_eval(ctx, None, &mut command);
-        ctx.editor.string_pool.release(command);
-
-        match result {
-            Ok(flow) => match flow {
-                EditorFlow::Continue => (),
-                _ => return flow,
-            },
-            Err(error) => {
-                ctx.editor
-                    .status_bar
-                    .write(MessageKind::Error)
-                    .fmt(format_args!(
-                        "{}:{}\n{}\n{}",
-                        config_name,
-                        line_index + 1,
-                        line,
-                        error
-                    ));
-                break;
-            }
-        }
-    }
-
-    EditorFlow::Continue
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
