@@ -1,13 +1,11 @@
-use std::process::Stdio;
-
 use crate::{
     buffer::BufferProperties,
     client::ClientHandle,
+    command::CommandManager,
     editor::{Editor, EditorContext, EditorFlow, KeysIterator},
-    editor_utils::{parse_process_command, MessageKind, ReadLine, ReadLinePoll},
+    editor_utils::{MessageKind, ReadLinePoll},
     mode::{ModeKind, ModeState},
-    picker::Picker,
-    platform::{Key, PlatformRequest, ProcessTag},
+    platform::{Key, KeyCode},
     word_database::WordIndicesIter,
 };
 
@@ -18,75 +16,14 @@ pub struct State {
         &mut KeysIterator,
         ReadLinePoll,
     ) -> Option<EditorFlow>,
-    find_file_waiting_for_process: bool,
-    find_file_buf: Vec<u8>,
-}
-
-impl State {
-    pub(crate) fn on_process_output(
-        &mut self,
-        picker: &mut Picker,
-        read_line: &ReadLine,
-        bytes: &[u8],
-    ) {
-        if !self.find_file_waiting_for_process {
-            return;
-        }
-
-        self.find_file_buf.extend_from_slice(bytes);
-
-        {
-            let mut filtered_entry_adder = picker.add_custom_filtered_entries(read_line.input());
-            if let Some(i) = self.find_file_buf.iter().rposition(|&b| b == b'\n') {
-                for line in self
-                    .find_file_buf
-                    .drain(..i + 1)
-                    .as_slice()
-                    .split(|&b| matches!(b, b'\n' | b'\r'))
-                {
-                    if line.is_empty() {
-                        continue;
-                    }
-                    if let Ok(line) = std::str::from_utf8(line) {
-                        filtered_entry_adder.add(line);
-                    }
-                }
-            }
-        }
-
-        picker.move_cursor(0);
-    }
-
-    pub(crate) fn on_process_exit(&mut self, picker: &mut Picker, read_line: &ReadLine) {
-        if !self.find_file_waiting_for_process {
-            return;
-        }
-
-        self.find_file_waiting_for_process = false;
-
-        {
-            let mut filtered_entry_adder = picker.add_custom_filtered_entries(read_line.input());
-            for line in self.find_file_buf.split(|&b| b == b'\n') {
-                if line.is_empty() {
-                    continue;
-                }
-                if let Ok(line) = std::str::from_utf8(line) {
-                    filtered_entry_adder.add(line);
-                }
-            }
-        }
-
-        self.find_file_buf.clear();
-        picker.move_cursor(0);
-    }
+    continuation: String,
 }
 
 impl Default for State {
     fn default() -> Self {
         Self {
             on_client_keys: |_, _, _, _| Some(EditorFlow::Continue),
-            find_file_waiting_for_process: false,
-            find_file_buf: Vec::new(),
+            continuation: String::new(),
         }
     }
 }
@@ -98,7 +35,6 @@ impl ModeState for State {
 
     fn on_exit(editor: &mut Editor) {
         editor.mode.plugin_handle = None;
-        editor.mode.picker_state.find_file_waiting_for_process = false;
         editor.read_line.input_mut().clear();
         editor.picker.clear();
     }
@@ -118,9 +54,46 @@ impl ModeState for State {
         if let ReadLinePoll::Pending = poll {
             keys.index = keys.index.saturating_sub(1);
             match keys.next(&ctx.editor.buffered_keys) {
-                Key::Ctrl('n') | Key::Down => ctx.editor.picker.move_cursor(1),
-                Key::Ctrl('p') | Key::Up => ctx.editor.picker.move_cursor(-1),
-                Key::Ctrl('j') | Key::PageDown => {
+                Key {
+                    code: KeyCode::Char('n'),
+                    shift: false,
+                    control: true,
+                    alt: false,
+                }
+                | Key {
+                    code: KeyCode::Down,
+                    shift: false,
+                    control: false,
+                    alt: false,
+                } => {
+                    ctx.editor.picker.move_cursor(1);
+                }
+                Key {
+                    code: KeyCode::Char('p'),
+                    shift: false,
+                    control: true,
+                    alt: false,
+                }
+                | Key {
+                    code: KeyCode::Up,
+                    shift: false,
+                    control: false,
+                    alt: false,
+                } => {
+                    ctx.editor.picker.move_cursor(-1);
+                }
+                Key {
+                    code: KeyCode::Char('j'),
+                    shift: false,
+                    control: true,
+                    alt: false,
+                }
+                | Key {
+                    code: KeyCode::PageDown,
+                    shift: false,
+                    control: false,
+                    alt: false,
+                } => {
                     let picker_height = ctx
                         .editor
                         .picker
@@ -129,7 +102,18 @@ impl ModeState for State {
                         as isize;
                     ctx.editor.picker.move_cursor(picker_height / 2);
                 }
-                Key::Ctrl('k') | Key::PageUp => {
+                Key {
+                    code: KeyCode::Char('k'),
+                    shift: false,
+                    control: true,
+                    alt: false,
+                }
+                | Key {
+                    code: KeyCode::PageUp,
+                    shift: false,
+                    control: false,
+                    alt: false,
+                } => {
                     let picker_height = ctx
                         .editor
                         .picker
@@ -138,11 +122,33 @@ impl ModeState for State {
                         as isize;
                     ctx.editor.picker.move_cursor(-picker_height / 2);
                 }
-                Key::Ctrl('b') | Key::Home => {
+                Key {
+                    code: KeyCode::Char('b'),
+                    shift: false,
+                    control: true,
+                    alt: false,
+                }
+                | Key {
+                    code: KeyCode::Home,
+                    shift: false,
+                    control: false,
+                    alt: false,
+                } => {
                     let cursor = ctx.editor.picker.cursor().unwrap_or(0) as isize;
                     ctx.editor.picker.move_cursor(-cursor);
                 }
-                Key::Ctrl('e') | Key::End => {
+                Key {
+                    code: KeyCode::Char('e'),
+                    shift: false,
+                    control: true,
+                    alt: false,
+                }
+                | Key {
+                    code: KeyCode::End,
+                    shift: false,
+                    control: false,
+                    alt: false,
+                } => {
                     let cursor = ctx.editor.picker.cursor().unwrap_or(0) as isize;
                     let entry_count = ctx.editor.picker.len() as isize;
                     ctx.editor.picker.move_cursor(entry_count - cursor - 1);
@@ -228,12 +234,10 @@ pub mod opened_buffers {
     }
 }
 
-pub mod find_file {
+pub mod custom {
     use super::*;
 
-    use std::path::Path;
-
-    pub fn enter_mode(ctx: &mut EditorContext, command: &str, prompt: &str) {
+    pub fn enter_mode(ctx: &mut EditorContext, continuation: &str, prompt: &str) {
         fn on_client_keys(
             ctx: &mut EditorContext,
             client_handle: ClientHandle,
@@ -241,76 +245,36 @@ pub mod find_file {
             poll: ReadLinePoll,
         ) -> Option<EditorFlow> {
             match poll {
-                ReadLinePoll::Pending => return Some(EditorFlow::Continue),
-                ReadLinePoll::Submitted => (),
-                ReadLinePoll::Canceled => {
+                ReadLinePoll::Pending => (),
+                ReadLinePoll::Submitted => {
+                    if ctx.editor.picker.cursor().is_none() {
+                        ctx.editor.enter_mode(ModeKind::default());
+                        return Some(EditorFlow::Continue);
+                    }
+
+                    let continuation = &ctx.editor.mode.picker_state.continuation;
+                    let continuation = ctx.editor.string_pool.acquire_with(continuation);
+                    let result = CommandManager::eval(ctx, Some(client_handle), &continuation);
+                    let flow = CommandManager::unwrap_eval_result(
+                        ctx,
+                        result,
+                        &continuation,
+                        Some("picker-continuation"),
+                    );
+                    ctx.editor.string_pool.release(continuation);
                     ctx.editor.enter_mode(ModeKind::default());
-                    return Some(EditorFlow::Continue);
+                    return Some(flow);
                 }
+                ReadLinePoll::Canceled => ctx.editor.enter_mode(ModeKind::default()),
             }
-
-            let path = match ctx.editor.picker.current_entry(&ctx.editor.word_database) {
-                Some((_, entry)) => entry,
-                _ => {
-                    ctx.editor.enter_mode(ModeKind::default());
-                    return Some(EditorFlow::Continue);
-                }
-            };
-
-            let path = ctx.editor.string_pool.acquire_with(path);
-            match ctx.editor.buffer_view_handle_from_path(
-                client_handle,
-                Path::new(&path),
-                BufferProperties::text(),
-                false,
-            ) {
-                Ok(buffer_view_handle) => {
-                    let client = ctx.clients.get_mut(client_handle);
-                    client
-                        .set_buffer_view_handle(Some(buffer_view_handle), &ctx.editor.buffer_views);
-                }
-                Err(error) => ctx
-                    .editor
-                    .status_bar
-                    .write(MessageKind::Error)
-                    .fmt(format_args!("{}", error)),
-            }
-            ctx.editor.string_pool.release(path);
-
-            ctx.editor.enter_mode(ModeKind::default());
             Some(EditorFlow::Continue)
         }
 
         ctx.editor.read_line.set_prompt(prompt);
-        ctx.editor.picker.clear();
-
-        let command = match parse_process_command(command) {
-            Some(mut command) => {
-                command.stdin(Stdio::null());
-                command.stdout(Stdio::piped());
-                command.stderr(Stdio::null());
-
-                command
-            }
-            None => {
-                ctx.editor
-                    .status_bar
-                    .write(MessageKind::Error)
-                    .fmt(format_args!("invalid find file command '{}'", command));
-                return;
-            }
-        };
-
-        ctx.editor.mode.picker_state.find_file_waiting_for_process = true;
-        ctx.platform
-            .requests
-            .enqueue(PlatformRequest::SpawnProcess {
-                tag: ProcessTag::FindFiles,
-                command,
-                buf_len: 4 * 1024,
-            });
-
-        ctx.editor.mode.picker_state.on_client_keys = on_client_keys;
+        let state = &mut ctx.editor.mode.picker_state;
+        state.on_client_keys = on_client_keys;
+        state.continuation.clear();
+        state.continuation.push_str(continuation);
         ctx.editor.enter_mode(ModeKind::Picker);
     }
 }
