@@ -1,80 +1,33 @@
-use std::process::Stdio;
-
 use crate::{
-    buffer::{BufferCollection, BufferHandle},
-    buffer_position::{BufferPosition, BufferPositionIndex},
+    buffer_position::BufferPositionIndex,
     buffer_view::CursorMovementKind,
     client::ClientHandle,
-    cursor::{Cursor, CursorCollection},
+    command::CommandManager,
+    cursor::{Cursor, CursorCollectionMutGuard},
     editor::{Editor, EditorContext, EditorFlow, KeysIterator},
-    editor_utils::{parse_process_command, MessageKind, ReadLinePoll, ResidualStrBytes},
-    events::EditorEventQueue,
+    editor_utils::{MessageKind, ReadLinePoll},
     mode::{ModeKind, ModeState},
     navigation_history::NavigationHistory,
     pattern::Pattern,
-    platform::{PlatformRequest, PooledBuf, ProcessTag},
-    word_database::WordDatabase,
 };
 
 pub struct State {
     pub on_client_keys:
         fn(&mut EditorContext, ClientHandle, &mut KeysIterator, ReadLinePoll) -> Option<EditorFlow>,
-    previous_position: BufferPosition,
-    find_pattern_command: String,
-    find_pattern_buffer_handle: Option<BufferHandle>,
-    find_pattern_residual_bytes: ResidualStrBytes,
-}
-
-impl State {
-    pub(crate) fn on_buffer_close(&mut self, buffer_handle: BufferHandle) {
-        if self.find_pattern_buffer_handle == Some(buffer_handle) {
-            self.find_pattern_buffer_handle = None;
-        }
-    }
-
-    pub(crate) fn on_process_output(
-        &mut self,
-        buffers: &mut BufferCollection,
-        word_database: &mut WordDatabase,
-        bytes: &[u8],
-        events: &mut EditorEventQueue,
-    ) {
-        if let Some(buffer_handle) = self.find_pattern_buffer_handle {
-            let mut buf = Default::default();
-            let texts = self
-                .find_pattern_residual_bytes
-                .receive_bytes(&mut buf, bytes);
-
-            let buffer = buffers.get_mut(buffer_handle);
-            for text in texts {
-                let position = buffer.content().end();
-                buffer.insert_text(word_database, position, text, events);
-            }
-        }
-    }
-
-    pub(crate) fn on_process_exit(
-        &mut self,
-        buffers: &mut BufferCollection,
-        word_database: &mut WordDatabase,
-        events: &mut EditorEventQueue,
-    ) {
-        self.on_process_output(buffers, word_database, &[], events);
-
-        self.find_pattern_command.clear();
-        self.find_pattern_buffer_handle = None;
-        self.find_pattern_residual_bytes = ResidualStrBytes::default();
-    }
+    previous_cursors: Vec<Cursor>,
+    previous_main_cursor_index: usize,
+    movement_kind: CursorMovementKind,
+    continuation: String,
 }
 
 impl Default for State {
     fn default() -> Self {
         Self {
             on_client_keys: |_, _, _, _| Some(EditorFlow::Continue),
-            previous_position: BufferPosition::zero(),
-            find_pattern_command: String::new(),
-            find_pattern_buffer_handle: None,
-            find_pattern_residual_bytes: ResidualStrBytes::default(),
+            previous_cursors: Vec::new(),
+            previous_main_cursor_index: 0,
+            movement_kind: CursorMovementKind::PositionAndAnchor,
+            continuation: String::new(),
         }
     }
 }
@@ -86,7 +39,6 @@ impl ModeState for State {
 
     fn on_exit(editor: &mut Editor) {
         editor.mode.plugin_handle = None;
-        editor.mode.read_line_state.find_pattern_command.clear();
         editor.read_line.input_mut().clear();
     }
 
@@ -111,7 +63,11 @@ pub mod search {
 
     use crate::editor_utils::SEARCH_REGISTER;
 
-    pub fn enter_mode(ctx: &mut EditorContext, client_handle: ClientHandle) {
+    pub fn enter_mode(
+        ctx: &mut EditorContext,
+        client_handle: ClientHandle,
+        movement_kind: CursorMovementKind,
+    ) {
         fn on_client_keys(
             ctx: &mut EditorContext,
             client_handle: ClientHandle,
@@ -132,32 +88,38 @@ pub mod search {
                         if search_ranges.is_empty() {
                             restore_saved_position(ctx, client_handle);
                         } else {
-                            let previous_position =
-                                ctx.editor.mode.read_line_state.previous_position;
-                            let current_position = buffer_view.cursors.main_cursor().position;
+                            let state = &ctx.editor.mode.read_line_state;
+                            let cursor_position = buffer_view.cursors.main_cursor().position;
 
                             {
                                 let mut cursors = buffer_view.cursors.mut_guard();
                                 cursors.clear();
-                                cursors.add(Cursor {
-                                    anchor: previous_position,
-                                    position: previous_position,
-                                });
+                                for &cursor in &state.previous_cursors {
+                                    cursors.add(cursor);
+                                }
+                                cursors.set_main_cursor_index(state.previous_main_cursor_index);
                             }
 
+                            let cursor_anchor = match state.movement_kind {
+                                CursorMovementKind::PositionAndAnchor => cursor_position,
+                                CursorMovementKind::PositionOnly => {
+                                    state.previous_cursors[state.previous_main_cursor_index].anchor
+                                }
+                            };
+
                             NavigationHistory::save_snapshot(client, &ctx.editor.buffer_views);
                             let buffer_view = ctx.editor.buffer_views.get_mut(buffer_view_handle);
                             {
                                 let mut cursors = buffer_view.cursors.mut_guard();
                                 cursors.clear();
                                 cursors.add(Cursor {
-                                    anchor: current_position,
-                                    position: current_position,
+                                    anchor: cursor_anchor,
+                                    position: cursor_position,
                                 });
                             }
 
                             ctx.editor.mode.normal_state.search_index = match search_ranges
-                                .binary_search_by_key(&current_position, |r| r.from)
+                                .binary_search_by_key(&cursor_position, |r| r.from)
                             {
                                 Ok(i) => i,
                                 Err(i) => i,
@@ -183,6 +145,7 @@ pub mod search {
         ctx.editor.read_line.set_prompt("search:");
         update_search(ctx, client_handle);
 
+        ctx.editor.mode.read_line_state.movement_kind = movement_kind;
         ctx.editor.mode.read_line_state.on_client_keys = on_client_keys;
         ctx.editor.enter_mode(ModeKind::ReadLine);
     }
@@ -348,29 +311,23 @@ pub mod filter_cursors {
         let mut cursors = buffer_view.cursors.mut_guard();
         let main_cursor_position = cursors.main_cursor().position;
 
-        let mut filtered_cursors = [Cursor::zero(); CursorCollection::capacity()];
-        let mut filtered_cursors_len = 0;
-
-        for &cursor in &cursors[..] {
-            if range_contains_pattern(buffer, cursor.to_range(), &ctx.editor.aux_pattern)
-                == keep_if_contains_pattern
+        for i in (0..cursors[..].len()).rev() {
+            let range = cursors[i].to_range();
+            if range_contains_pattern(buffer, range, &ctx.editor.aux_pattern)
+                != keep_if_contains_pattern
             {
-                filtered_cursors[filtered_cursors_len] = cursor;
-                filtered_cursors_len += 1;
+                cursors.swap_remove(i);
             }
         }
 
-        cursors.clear();
-        for &cursor in &filtered_cursors[..filtered_cursors_len] {
-            cursors.add(cursor);
-        }
-
         if cursors[..].is_empty() {
             cursors.add(Cursor {
                 anchor: main_cursor_position,
                 position: main_cursor_position,
             });
         }
+
+        cursors.set_main_cursor_near_position(main_cursor_position);
     }
 }
 
@@ -381,12 +338,11 @@ pub mod split_cursors {
 
     pub fn enter_by_pattern_mode(ctx: &mut EditorContext) {
         fn add_matches(
-            cursors: &mut [Cursor],
-            mut cursors_len: usize,
+            cursors: &mut CursorCollectionMutGuard,
             line: &str,
             pattern: &Pattern,
             start_position: BufferPosition,
-        ) -> usize {
+        ) {
             let search_anchor = pattern.search_anchor();
             for range in pattern.match_indices(line, search_anchor) {
                 let mut anchor = start_position;
@@ -394,14 +350,8 @@ pub mod split_cursors {
                 let mut position = start_position;
                 position.column_byte_index += range.end as BufferPositionIndex;
 
-                if cursors_len >= cursors.len() {
-                    return cursors.len();
-                }
-                cursors[cursors_len] = Cursor { anchor, position };
-                cursors_len += 1;
+                cursors.add(Cursor { anchor, position });
             }
-
-            cursors_len
         }
 
         ctx.editor.read_line.set_prompt("split-by:");
@@ -416,12 +366,11 @@ pub mod split_cursors {
 
     pub fn enter_by_separators_mode(ctx: &mut EditorContext) {
         fn add_matches(
-            cursors: &mut [Cursor],
-            mut cursors_len: usize,
+            cursors: &mut CursorCollectionMutGuard,
             line: &str,
             pattern: &Pattern,
             start_position: BufferPosition,
-        ) -> usize {
+        ) {
             let search_anchor = pattern.search_anchor();
             let mut index = 0;
             for range in pattern.match_indices(line, search_anchor) {
@@ -431,21 +380,14 @@ pub mod split_cursors {
                     let mut position = start_position;
                     position.column_byte_index += range.start as BufferPositionIndex;
 
-                    if cursors_len >= cursors.len() {
-                        return cursors.len();
-                    }
-                    cursors[cursors_len] = Cursor { anchor, position };
-                    cursors_len += 1;
+                    cursors.add(Cursor { anchor, position });
                 }
 
                 index = range.end;
             }
 
             if index < line.len() {
-                if cursors_len >= cursors.len() {
-                    return cursors.len();
-                }
-                cursors[cursors_len] = Cursor {
+                cursors.add(Cursor {
                     anchor: BufferPosition::line_col(
                         start_position.line_index,
                         start_position.column_byte_index + index as BufferPositionIndex,
@@ -454,11 +396,8 @@ pub mod split_cursors {
                         start_position.line_index,
                         start_position.column_byte_index + line.len() as BufferPositionIndex,
                     ),
-                };
-                cursors_len += 1;
+                });
             }
-
-            cursors_len
         }
 
         ctx.editor.read_line.set_prompt("split-on:");
@@ -474,7 +413,7 @@ pub mod split_cursors {
     fn on_event_impl(
         ctx: &mut EditorContext,
         client_handle: ClientHandle,
-        add_matches: fn(&mut [Cursor], usize, &str, &Pattern, BufferPosition) -> usize,
+        add_matches: fn(&mut CursorCollectionMutGuard, &str, &Pattern, BufferPosition),
     ) {
         let pattern = ctx.editor.read_line.input();
         let pattern = if pattern.is_empty() {
@@ -505,39 +444,27 @@ pub mod split_cursors {
         let mut cursors = buffer_view.cursors.mut_guard();
         let main_cursor_position = cursors.main_cursor().position;
 
-        let mut splitted_cursors = [Cursor::zero(); CursorCollection::capacity()];
-        let mut splitted_cursors_len = 0;
-
-        for cursor in &cursors[..] {
+        let mut i = cursors[..].len() - 1;
+        loop {
+            let cursor = cursors.swap_remove(i);
             let range = cursor.to_range();
-            let new_cursors_start_index = splitted_cursors_len;
+            let new_cursors_start_index = cursors[..].len();
+
+            let main_cursor_index = cursors.main_cursor_index();
 
             if range.from.line_index == range.to.line_index {
                 let line = &buffer.lines()[range.from.line_index as usize].as_str()
                     [range.from.column_byte_index as usize..range.to.column_byte_index as usize];
-                splitted_cursors_len = add_matches(
-                    &mut splitted_cursors,
-                    splitted_cursors_len,
-                    line,
-                    &ctx.editor.aux_pattern,
-                    range.from,
-                );
+                add_matches(&mut cursors, line, &ctx.editor.aux_pattern, range.from);
             } else {
                 let line = &buffer.lines()[range.from.line_index as usize].as_str()
                     [range.from.column_byte_index as usize..];
-                splitted_cursors_len = add_matches(
-                    &mut splitted_cursors,
-                    splitted_cursors_len,
-                    line,
-                    &ctx.editor.aux_pattern,
-                    range.from,
-                );
+                add_matches(&mut cursors, line, &ctx.editor.aux_pattern, range.from);
 
                 for line_index in (range.from.line_index + 1)..range.to.line_index {
                     let line = buffer.lines()[line_index as usize].as_str();
-                    splitted_cursors_len = add_matches(
-                        &mut splitted_cursors,
-                        splitted_cursors_len,
+                    add_matches(
+                        &mut cursors,
                         line,
                         &ctx.editor.aux_pattern,
                         BufferPosition::line_col(line_index, 0),
@@ -546,9 +473,8 @@ pub mod split_cursors {
 
                 let line = &buffer.lines()[range.to.line_index as usize].as_str()
                     [..range.to.column_byte_index as usize];
-                splitted_cursors_len = add_matches(
-                    &mut splitted_cursors,
-                    splitted_cursors_len,
+                add_matches(
+                    &mut cursors,
                     line,
                     &ctx.editor.aux_pattern,
                     BufferPosition::line_col(range.to.line_index, 0),
@@ -556,15 +482,20 @@ pub mod split_cursors {
             }
 
             if cursor.position == range.from {
-                for cursor in &mut splitted_cursors[new_cursors_start_index..] {
+                for cursor in &mut cursors[new_cursors_start_index..] {
                     std::mem::swap(&mut cursor.anchor, &mut cursor.position);
                 }
             }
-        }
 
-        cursors.clear();
-        for &cursor in &splitted_cursors[..splitted_cursors_len] {
-            cursors.add(cursor);
+            if i < main_cursor_index {
+                let new_cursor_count = cursors[..].len() - new_cursors_start_index;
+                cursors.set_main_cursor_index(main_cursor_index + new_cursor_count);
+            }
+
+            if i == 0 {
+                break;
+            }
+            i -= 1;
         }
 
         if cursors[..].is_empty() {
@@ -573,6 +504,8 @@ pub mod split_cursors {
                 position: main_cursor_position,
             });
         }
+
+        cursors.set_main_cursor_near_position(main_cursor_position);
     }
 }
 
@@ -631,10 +564,10 @@ pub mod goto {
     }
 }
 
-pub mod process {
+pub mod custom {
     use super::*;
 
-    pub fn enter_replace_mode(ctx: &mut EditorContext) {
+    pub fn enter_mode(ctx: &mut EditorContext, continuation: &str, prompt: &str) {
         fn on_client_keys(
             ctx: &mut EditorContext,
             client_handle: ClientHandle,
@@ -642,263 +575,31 @@ pub mod process {
             poll: ReadLinePoll,
         ) -> Option<EditorFlow> {
             match poll {
-                ReadLinePoll::Pending => Some(EditorFlow::Continue),
+                ReadLinePoll::Pending => (),
                 ReadLinePoll::Submitted => {
-                    spawn_process(ctx, client_handle, true);
+                    let continuation = &ctx.editor.mode.read_line_state.continuation;
+                    let continuation = ctx.editor.string_pool.acquire_with(continuation);
+                    let result = CommandManager::eval(ctx, Some(client_handle), &continuation);
+                    let flow = CommandManager::unwrap_eval_result(
+                        ctx,
+                        result,
+                        &continuation,
+                        Some("readline-continuation"),
+                    );
+                    ctx.editor.string_pool.release(continuation);
                     ctx.editor.enter_mode(ModeKind::default());
-                    Some(EditorFlow::Continue)
-                }
-                ReadLinePoll::Canceled => {
-                    ctx.editor.enter_mode(ModeKind::default());
-                    Some(EditorFlow::Continue)
+                    return Some(flow);
                 }
+                ReadLinePoll::Canceled => ctx.editor.enter_mode(ModeKind::default()),
             }
-        }
-
-        ctx.editor.read_line.set_prompt("replace-with-output:");
-        ctx.editor.mode.read_line_state.on_client_keys = on_client_keys;
-        ctx.editor.enter_mode(ModeKind::ReadLine);
-    }
-
-    pub fn enter_insert_mode(ctx: &mut EditorContext) {
-        fn on_client_keys(
-            ctx: &mut EditorContext,
-            client_handle: ClientHandle,
-            _: &mut KeysIterator,
-            poll: ReadLinePoll,
-        ) -> Option<EditorFlow> {
-            match poll {
-                ReadLinePoll::Pending => Some(EditorFlow::Continue),
-                ReadLinePoll::Submitted => {
-                    spawn_process(ctx, client_handle, false);
-                    ctx.editor.enter_mode(ModeKind::default());
-                    Some(EditorFlow::Continue)
-                }
-                ReadLinePoll::Canceled => {
-                    ctx.editor.enter_mode(ModeKind::default());
-                    Some(EditorFlow::Continue)
-                }
-            }
-        }
-
-        ctx.editor.read_line.set_prompt("insert-from-output:");
-        ctx.editor.mode.read_line_state.on_client_keys = on_client_keys;
-        ctx.editor.enter_mode(ModeKind::ReadLine);
-    }
-
-    pub fn enter_run_mode(ctx: &mut EditorContext) {
-        fn on_client_keys(
-            ctx: &mut EditorContext,
-            _: ClientHandle,
-            _: &mut KeysIterator,
-            poll: ReadLinePoll,
-        ) -> Option<EditorFlow> {
-            match poll {
-                ReadLinePoll::Pending => Some(EditorFlow::Continue),
-                ReadLinePoll::Submitted => {
-                    let command = ctx.editor.read_line.input();
-                    if let Some(mut command) = parse_process_command(command) {
-                        command.stdin(Stdio::null());
-                        command.stdout(Stdio::null());
-                        command.stderr(Stdio::null());
-
-                        ctx.platform
-                            .requests
-                            .enqueue(PlatformRequest::SpawnProcess {
-                                tag: ProcessTag::Ignored,
-                                command,
-                                buf_len: 0,
-                            });
-                    }
-
-                    ctx.editor.enter_mode(ModeKind::default());
-                    Some(EditorFlow::Continue)
-                }
-                ReadLinePoll::Canceled => {
-                    ctx.editor.enter_mode(ModeKind::default());
-                    Some(EditorFlow::Continue)
-                }
-            }
-        }
-
-        ctx.editor.read_line.set_prompt("run-command:");
-        ctx.editor.mode.read_line_state.on_client_keys = on_client_keys;
-        ctx.editor.enter_mode(ModeKind::ReadLine);
-    }
-
-    fn spawn_process(ctx: &mut EditorContext, client_handle: ClientHandle, pipe: bool) {
-        let buffer_view_handle = match ctx.clients.get(client_handle).buffer_view_handle() {
-            Some(handle) => handle,
-            None => return,
-        };
-        let buffer_view = ctx.editor.buffer_views.get_mut(buffer_view_handle);
-        let content = ctx.editor.buffers.get(buffer_view.buffer_handle).content();
-
-        const NONE_POOLED_BUF: Option<PooledBuf> = None;
-        let mut stdins = [NONE_POOLED_BUF; CursorCollection::capacity()];
-
-        if pipe {
-            for (i, cursor) in buffer_view.cursors[..].iter().enumerate() {
-                let range = cursor.to_range();
-
-                let mut buf = ctx.platform.buf_pool.acquire();
-                let write = buf.write();
-                for text in content.text_range(range) {
-                    write.extend_from_slice(text.as_bytes());
-                }
-
-                stdins[i] = Some(buf);
-            }
-        }
-
-        buffer_view.delete_text_in_cursor_ranges(
-            &mut ctx.editor.buffers,
-            &mut ctx.editor.word_database,
-            &mut ctx.editor.events,
-        );
-
-        ctx.trigger_event_handlers();
-
-        let command = ctx.editor.read_line.input();
-        let buffer_view = ctx.editor.buffer_views.get_mut(buffer_view_handle);
-        for (i, cursor) in buffer_view.cursors[..].iter().enumerate() {
-            let command = match parse_process_command(command) {
-                Some(command) => command,
-                None => continue,
-            };
-
-            ctx.editor.buffers.spawn_insert_process(
-                &mut ctx.platform,
-                command,
-                buffer_view.buffer_handle,
-                cursor.position,
-                stdins[i].take(),
-            );
-        }
-    }
-}
-
-pub mod find_pattern {
-    use super::*;
-
-    use std::{path::Path, process::Stdio};
-
-    use crate::{
-        buffer::BufferProperties,
-        buffer_position::BufferRange,
-        platform::{PlatformRequest, ProcessTag},
-    };
-
-    pub fn enter_mode(ctx: &mut EditorContext, command: &str, prompt: &str) {
-        fn on_client_keys(
-            ctx: &mut EditorContext,
-            client_handle: ClientHandle,
-            _: &mut KeysIterator,
-            poll: ReadLinePoll,
-        ) -> Option<EditorFlow> {
-            match poll {
-                ReadLinePoll::Pending => return Some(EditorFlow::Continue),
-                ReadLinePoll::Submitted => (),
-                ReadLinePoll::Canceled => {
-                    ctx.editor.enter_mode(ModeKind::default());
-                    return Some(EditorFlow::Continue);
-                }
-            }
-
-            let mut buffer_name = ctx.editor.string_pool.acquire();
-            buffer_name.push_str(ctx.editor.read_line.input());
-            buffer_name.push_str(".refs");
-            let buffer_view_handle = ctx.editor.buffer_view_handle_from_path(
-                client_handle,
-                Path::new(&buffer_name),
-                BufferProperties::scratch(),
-                true,
-            );
-            ctx.editor.string_pool.release(buffer_name);
-
-            let buffer_view_handle = match buffer_view_handle {
-                Ok(handle) => handle,
-                Err(error) => {
-                    ctx.editor
-                        .status_bar
-                        .write(MessageKind::Error)
-                        .fmt(format_args!("{}", error));
-                    return Some(EditorFlow::Continue);
-                }
-            };
-
-            let buffer_view = ctx.editor.buffer_views.get_mut(buffer_view_handle);
-            let buffer = ctx.editor.buffers.get_mut(buffer_view.buffer_handle);
-
-            buffer.properties = BufferProperties::scratch();
-            let range = BufferRange::between(BufferPosition::zero(), buffer.content().end());
-            buffer.delete_range(&mut ctx.editor.word_database, range, &mut ctx.editor.events);
-
-            let state = &mut ctx.editor.mode.read_line_state;
-            state.find_pattern_buffer_handle = Some(buffer.handle());
-            state.find_pattern_residual_bytes = ResidualStrBytes::default();
-
-            const REPLACE_PATTERN: &str = "{}";
-            if let Some(i) = state.find_pattern_command.find(REPLACE_PATTERN) {
-                state
-                    .find_pattern_command
-                    .replace_range(i..i + REPLACE_PATTERN.len(), ctx.editor.read_line.input());
-            }
-
-            let command = match parse_process_command(&state.find_pattern_command) {
-                Some(mut command) => {
-                    command.stdin(Stdio::null());
-                    command.stdout(Stdio::piped());
-                    command.stderr(Stdio::null());
-
-                    command
-                }
-                None => {
-                    ctx.editor
-                        .status_bar
-                        .write(MessageKind::Error)
-                        .fmt(format_args!(
-                            "invalid find pattern command '{}'",
-                            &state.find_pattern_command
-                        ));
-                    return Some(EditorFlow::Continue);
-                }
-            };
-
-            ctx.platform
-                .requests
-                .enqueue(PlatformRequest::SpawnProcess {
-                    tag: ProcessTag::FindPattern,
-                    command,
-                    buf_len: 4 * 1024,
-                });
-
-            let client = ctx.clients.get_mut(client_handle);
-            client.set_buffer_view_handle(Some(buffer_view_handle), &ctx.editor.buffer_views);
-
-            {
-                let mut cursors = ctx
-                    .editor
-                    .buffer_views
-                    .get_mut(buffer_view_handle)
-                    .cursors
-                    .mut_guard();
-                cursors.clear();
-                cursors.add(Cursor {
-                    anchor: BufferPosition::zero(),
-                    position: BufferPosition::zero(),
-                });
-            }
-
-            ctx.editor.enter_mode(ModeKind::default());
             Some(EditorFlow::Continue)
         }
 
         ctx.editor.read_line.set_prompt(prompt);
         let state = &mut ctx.editor.mode.read_line_state;
         state.on_client_keys = on_client_keys;
-        state.find_pattern_command.clear();
-        state.find_pattern_command.push_str(command);
+        state.continuation.clear();
+        state.continuation.push_str(continuation);
         ctx.editor.enter_mode(ModeKind::ReadLine);
     }
 }
@@ -909,7 +610,12 @@ fn save_current_position(ctx: &mut EditorContext, client_handle: ClientHandle) {
         None => return,
     };
     let buffer_view = ctx.editor.buffer_views.get(buffer_view_handle);
-    ctx.editor.mode.read_line_state.previous_position = buffer_view.cursors.main_cursor().position;
+    let state = &mut ctx.editor.mode.read_line_state;
+    state.previous_cursors.clear();
+    for &cursor in &buffer_view.cursors[..] {
+        state.previous_cursors.push(cursor);
+    }
+    state.previous_main_cursor_index = buffer_view.cursors.main_cursor_index();
 }
 
 fn restore_saved_position(ctx: &mut EditorContext, client_handle: ClientHandle) {
@@ -917,12 +623,13 @@ fn restore_saved_position(ctx: &mut EditorContext, client_handle: ClientHandle)
         Some(handle) => handle,
         None => return,
     };
-    let position = ctx.editor.mode.read_line_state.previous_position;
     let buffer_view = ctx.editor.buffer_views.get_mut(buffer_view_handle);
     let mut cursors = buffer_view.cursors.mut_guard();
     cursors.clear();
-    cursors.add(Cursor {
-        anchor: position,
-        position,
-    });
+
+    let state = &ctx.editor.mode.read_line_state;
+    for &cursor in &state.previous_cursors {
+        cursors.add(cursor);
+    }
+    cursors.set_main_cursor_index(state.previous_main_cursor_index);
 }
