@@ -7,10 +7,9 @@ use std::{
 use crate::{client::ClientHandle, editor_utils::parse_process_command, plugin::PluginHandle};
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
-pub enum Key {
+pub enum KeyCode {
     None,
     Backspace,
-    Enter,
     Left,
     Right,
     Up,
@@ -19,15 +18,31 @@ pub enum Key {
     End,
     PageUp,
     PageDown,
-    Tab,
     Delete,
     F(u8),
     Char(char),
-    Ctrl(char),
-    Alt(char),
     Esc,
 }
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub struct Key {
+    pub code: KeyCode,
+    pub shift: bool,
+    pub control: bool,
+    pub alt: bool,
+}
+
+impl Default for Key {
+    fn default() -> Self {
+        Self {
+            code: KeyCode::None,
+            shift: false,
+            control: false,
+            alt: false,
+        }
+    }
+}
+
 pub enum PlatformEvent {
     Idle,
     ConnectionOpen {
@@ -84,8 +99,7 @@ pub enum PlatformRequest {
 pub enum ProcessTag {
     Ignored,
     Buffer(u32),
-    FindFiles,
-    FindPattern,
+    PickerEntries,
     Plugin {
         plugin_handle: PluginHandle,
         id: u32,
@@ -139,7 +153,6 @@ impl Platform {
             command.stderr(Stdio::null());
             if let Ok(output) = command.output() {
                 if let Ok(output) = String::from_utf8(output.stdout) {
-                    text.clear();
                     text.push_str(&output);
                 }
             }
