@@ -14,10 +14,88 @@ use std::{
 use crate::{
     application::{ApplicationConfig, ClientApplication},
     editor_utils::hash_bytes,
-    platform::{BufPool, Key, PooledBuf, ProcessTag},
+    platform::{BufPool, Key, KeyCode, PooledBuf, ProcessTag},
     Args,
 };
 
+fn spawn_server() {
+    let mut file_actions = unsafe {
+        let mut file_actions = std::mem::zeroed::<libc::posix_spawn_file_actions_t>();
+        if libc::posix_spawn_file_actions_init(&mut file_actions) != 0 {
+            panic!("could not init posix spawn file actions");
+        }
+        if libc::posix_spawn_file_actions_addclose(&mut file_actions, libc::STDIN_FILENO) != 0 {
+            panic!("could not add close stdin to posix spawn file actions");
+        }
+        if libc::posix_spawn_file_actions_addclose(&mut file_actions, libc::STDOUT_FILENO) != 0 {
+            panic!("could not add close stdout to posix spawn file actions");
+        }
+        file_actions
+    };
+
+    let mut attributes = unsafe {
+        let mut attributes = std::mem::zeroed::<libc::posix_spawnattr_t>();
+        if libc::posix_spawnattr_init(&mut attributes) != 0 {
+            panic!("could not init posix spawn attributes");
+        }
+        if libc::posix_spawnattr_setflags(&mut attributes, libc::POSIX_SPAWN_SETPGROUP as _) != 0 {
+            panic!("could not set posix spawn attributes flags");
+        }
+        if libc::posix_spawnattr_setpgroup(&mut attributes, 0) != 0 {
+            panic!("could not set pgroup to posix spawn attributes");
+        }
+        attributes
+    };
+
+    let argv_owned: Vec<_> = std::env::args_os().collect();
+    let mut argv = Vec::new();
+    let mut args = argv_owned.iter();
+    match args.next() {
+        Some(arg) => argv.push(arg.as_bytes().as_ptr()),
+        None => panic!("could not extract process path from argv"),
+    }
+    argv.push("--server\0".as_ptr());
+    for arg in args {
+        argv.push(arg.as_bytes().as_ptr());
+    }
+    argv.push(std::ptr::null());
+
+    let mut envp_owned = Vec::new();
+    for (key, value) in std::env::vars_os() {
+        let mut env = key;
+        env.push("=");
+        env.push(&value);
+        env.push("\0");
+        envp_owned.push(env);
+    }
+    let mut envp = Vec::new();
+    for var in &envp_owned {
+        envp.push(var.as_bytes().as_ptr());
+    }
+    envp.push(std::ptr::null());
+
+    unsafe {
+        let result = libc::posix_spawn(
+            std::ptr::null_mut(),
+            argv[0] as _,
+            &file_actions,
+            &attributes,
+            argv.as_ptr() as _,
+            envp.as_ptr() as _,
+        );
+        if result != 0 {
+            panic!("could not spawn server");
+        }
+
+        if libc::posix_spawn_file_actions_destroy(&mut file_actions) != 0 {
+            panic!("could not destroy posix spawn file actions");
+        }
+        if libc::posix_spawnattr_destroy(&mut attributes) != 0 {
+            panic!("could not destroy posix spawn attributes");
+        }
+    }
+}
+
 pub(crate) fn run(
     config: ApplicationConfig,
     server_fn: fn(ApplicationConfig, UnixListener),
@@ -70,14 +148,9 @@ pub(crate) fn run(
     } else {
         match UnixStream::connect(session_path) {
             Ok(stream) => client_fn(config.args, stream),
-            Err(_) => match unsafe { libc::fork() } {
-                -1 => panic!("could not start server"),
-                0 => {
-                    unsafe { libc::daemon(true as _, false as _) };
-                    server_fn(config, start_server(session_path));
-                    let _ = fs::remove_file(session_path);
-                }
-                _ => loop {
+            Err(_) => {
+                spawn_server();
+                loop {
                     match UnixStream::connect(session_path) {
                         Ok(stream) => {
                             client_fn(config.args, stream);
@@ -85,8 +158,8 @@ pub(crate) fn run(
                         }
                         Err(_) => std::thread::sleep(Duration::from_millis(100)),
                     }
-                },
-            },
+                }
+            }
         }
     }
 }
@@ -138,9 +211,16 @@ impl Terminal {
         next_state.c_cc[libc::VMIN] = 0;
         next_state.c_cc[libc::VTIME] = 0;
         unsafe { libc::tcsetattr(self.fd, libc::TCSANOW, &next_state) };
+
+        // TODO: enable kitty keyboard protocol
+        // https://sw.kovidgoyal.net/kitty/keyboard-protocol/
+        //write_all_bytes(self.fd, b"\x1b[>1u");
     }
 
     pub fn leave_raw_mode(&self) {
+        // TODO: enable kitty keyboard protocol
+        // https://sw.kovidgoyal.net/kitty/keyboard-protocol/
+        //write_all_bytes(self.fd, b"\x1b[<u");
         unsafe { libc::tcsetattr(self.fd, libc::TCSAFLUSH, &self.original_state) };
     }
 
@@ -163,47 +243,71 @@ impl Terminal {
     pub fn parse_keys(&self, mut buf: &[u8], keys: &mut Vec<Key>) {
         let backspace_code = self.original_state.c_cc[libc::VERASE];
         loop {
-            let (key, rest) = match buf {
+            let mut shift = false;
+            let mut control = false;
+            let alt = false;
+
+            let (mut code, rest) = match buf {
                 &[] => break,
-                &[b, ref rest @ ..] if b == backspace_code => (Key::Backspace, rest),
-                &[0x1b, b'[', b'5', b'~', ref rest @ ..] => (Key::PageUp, rest),
-                &[0x1b, b'[', b'6', b'~', ref rest @ ..] => (Key::PageDown, rest),
-                &[0x1b, b'[', b'A', ref rest @ ..] => (Key::Up, rest),
-                &[0x1b, b'[', b'B', ref rest @ ..] => (Key::Down, rest),
-                &[0x1b, b'[', b'C', ref rest @ ..] => (Key::Right, rest),
-                &[0x1b, b'[', b'D', ref rest @ ..] => (Key::Left, rest),
+                &[b, ref rest @ ..] if b == backspace_code => (KeyCode::Backspace, rest),
+                &[0x1b, b'[', b'5', b'~', ref rest @ ..] => (KeyCode::PageUp, rest),
+                &[0x1b, b'[', b'6', b'~', ref rest @ ..] => (KeyCode::PageDown, rest),
+                &[0x1b, b'[', b'A', ref rest @ ..] => (KeyCode::Up, rest),
+                &[0x1b, b'[', b'B', ref rest @ ..] => (KeyCode::Down, rest),
+                &[0x1b, b'[', b'C', ref rest @ ..] => (KeyCode::Right, rest),
+                &[0x1b, b'[', b'D', ref rest @ ..] => (KeyCode::Left, rest),
+                &[0x1b, b'[', b'1', b'3', b'u', ref rest @ ..] => (KeyCode::Char('\n'), rest),
+                &[0x1b, b'[', b'2', b'7', b'u', ref rest @ ..] => (KeyCode::Esc, rest),
                 &[0x1b, b'[', b'1', b'~', ref rest @ ..]
                 | &[0x1b, b'[', b'7', b'~', ref rest @ ..]
                 | &[0x1b, b'[', b'H', ref rest @ ..]
-                | &[0x1b, b'O', b'H', ref rest @ ..] => (Key::Home, rest),
+                | &[0x1b, b'O', b'H', ref rest @ ..] => (KeyCode::Home, rest),
                 &[0x1b, b'[', b'4', b'~', ref rest @ ..]
                 | &[0x1b, b'[', b'8', b'~', ref rest @ ..]
                 | &[0x1b, b'[', b'F', ref rest @ ..]
-                | &[0x1b, b'O', b'F', ref rest @ ..] => (Key::End, rest),
-                &[0x1b, b'[', b'3', b'~', ref rest @ ..] => (Key::Delete, rest),
-                &[0x1b, ref rest @ ..] => (Key::Esc, rest),
-                &[0x8, ref rest @ ..] => (Key::Backspace, rest),
-                &[b'\r', ref rest @ ..] => (Key::Enter, rest),
-                &[b'\t', ref rest @ ..] => (Key::Tab, rest),
-                &[0x7f, ref rest @ ..] => (Key::Delete, rest),
+                | &[0x1b, b'O', b'F', ref rest @ ..] => (KeyCode::End, rest),
+                &[0x1b, b'[', b'3', b'~', ref rest @ ..] => (KeyCode::Delete, rest),
+                &[0x1b, b'[', b'9', b'u', ref rest @ ..] => (KeyCode::Char('\t'), rest),
+                &[0x1b, ref rest @ ..] => (KeyCode::Esc, rest),
+                &[0x8, ref rest @ ..] => (KeyCode::Backspace, rest),
+                &[b'\r', ref rest @ ..] => (KeyCode::Char('\n'), rest),
+                &[b'\t', ref rest @ ..] => (KeyCode::Char('\t'), rest),
+                &[0x7f, ref rest @ ..] => (KeyCode::Delete, rest),
                 &[b @ 0b0..=0b11111, ref rest @ ..] => {
+                    control = true;
                     let byte = b | 0b01100000;
-                    (Key::Ctrl(byte as _), rest)
+                    (KeyCode::Char(byte as _), rest)
                 }
                 _ => match buf.iter().position(|b| b.is_ascii()).unwrap_or(buf.len()) {
-                    0 => (Key::Char(buf[0] as _), &buf[1..]),
+                    0 => (KeyCode::Char(buf[0] as _), &buf[1..]),
                     len => {
                         let (c, rest) = buf.split_at(len);
                         match std::str::from_utf8(c) {
                             Ok(s) => match s.chars().next() {
-                                Some(c) => (Key::Char(c), rest),
-                                None => (Key::None, rest),
+                                Some(c) => (KeyCode::Char(c), rest),
+                                None => (KeyCode::None, rest),
                             },
-                            Err(_) => (Key::None, rest),
+                            Err(_) => (KeyCode::None, rest),
                         }
                     }
                 },
             };
+
+            if let KeyCode::Char(c) = &mut code {
+                if shift {
+                    *c = c.to_ascii_uppercase();
+                } else {
+                    shift = c.is_ascii_uppercase();
+                }
+            }
+
+            let key = Key {
+                code,
+                shift,
+                control,
+                alt,
+            };
+
             buf = rest;
             keys.push(key);
         }
