@@ -48,7 +48,7 @@ use winapi::{
         },
         wincontypes::{
             INPUT_RECORD, KEY_EVENT, LEFT_ALT_PRESSED, LEFT_CTRL_PRESSED, RIGHT_ALT_PRESSED,
-            RIGHT_CTRL_PRESSED, WINDOW_BUFFER_SIZE_EVENT,
+            RIGHT_CTRL_PRESSED, SHIFT_PRESSED, WINDOW_BUFFER_SIZE_EVENT,
         },
         winnls::CP_UTF8,
         winnt::{
@@ -71,7 +71,7 @@ use crate::{
     client::ClientHandle,
     editor_utils::hash_bytes,
     platform::{
-        drop_request, BufPool, Key, PlatformEvent, PlatformProcessHandle, PlatformRequest,
+        drop_request, BufPool, Key, KeyCode, PlatformEvent, PlatformProcessHandle, PlatformRequest,
         PooledBuf, ProcessTag,
     },
     Args,
@@ -181,11 +181,13 @@ pub fn main(config: ApplicationConfig) {
 
     pipe_path.clear();
     pipe_path.extend(PIPE_PREFIX.encode_utf16());
+    pipe_path.extend(env!("CARGO_PKG_NAME").encode_utf16());
+    pipe_path.push(b'-' as _);
     pipe_path.extend(session_name.encode_utf16());
-    pipe_path.push(0);
+    pipe_path.push(b'\0' as _);
 
     if config.args.print_session {
-        print!("{}{}", PIPE_PREFIX, session_name);
+        print!("{}{}-{}", PIPE_PREFIX, env!("CARGO_PKG_NAME"), session_name);
         return;
     }
 
@@ -195,7 +197,7 @@ pub fn main(config: ApplicationConfig) {
         }
     } else {
         if !pipe_exists(&pipe_path) {
-            fork();
+            spawn_server();
             while !pipe_exists(&pipe_path) {
                 std::thread::sleep(Duration::from_millis(100));
             }
@@ -454,7 +456,7 @@ fn global_unlock(handle: HANDLE) {
     unsafe { GlobalUnlock(handle) };
 }
 
-fn fork() {
+fn spawn_server() {
     let mut startup_info = unsafe { std::mem::zeroed::<STARTUPINFOW>() };
     startup_info.cb = std::mem::size_of::<STARTUPINFOW>() as _;
     startup_info.dwFlags = STARTF_USESTDHANDLES;
@@ -464,19 +466,44 @@ fn fork() {
 
     let mut process_info = unsafe { std::mem::zeroed::<PROCESS_INFORMATION>() };
 
-    let mut client_command_line = unsafe { GetCommandLineW() };
-    let mut command_line = Vec::with_capacity(1024);
-    loop {
-        unsafe {
-            let short = std::ptr::read(client_command_line);
-            if short == 0 {
+    let client_command_line = unsafe {
+        let command_line_start = GetCommandLineW();
+        let mut command_line_end = command_line_start;
+        loop {
+            if std::ptr::read(command_line_end) == 0 {
                 break;
             }
-            client_command_line = client_command_line.offset(1);
-            command_line.push(short);
+            command_line_end = command_line_end.offset(1);
         }
+        let len = command_line_end.offset_from(command_line_start);
+        std::slice::from_raw_parts(command_line_start, len as _)
+    };
+
+    if client_command_line.is_empty() {
+        panic!("executable command line was empty");
     }
-    command_line.extend_from_slice(&b" --server".map(|b| b as _));
+    let application_name_len = if client_command_line[0] == b'"' as _ {
+        match client_command_line[1..]
+            .iter()
+            .position(|&s| s == b'"' as _)
+        {
+            Some(i) => i + 2,
+            None => client_command_line.len(),
+        }
+    } else {
+        match client_command_line.iter().position(|&s| s == b' ' as _) {
+            Some(i) => i,
+            None => client_command_line.len(),
+        }
+    };
+    let (client_application_name, client_command_line) =
+        client_command_line.split_at(application_name_len);
+
+    let server_flag = b" --server";
+    let mut command_line = Vec::with_capacity(client_command_line.len() + server_flag.len());
+    command_line.extend_from_slice(client_application_name);
+    command_line.extend_from_slice(&server_flag.map(|b| b as _));
+    command_line.extend_from_slice(client_command_line);
     command_line.push(0);
 
     let result = unsafe {
@@ -1057,7 +1084,6 @@ struct EventListener {
 impl EventListener {
     pub fn new() -> Self {
         const DEFAULT_EVENT_SOURCE: EventSource = EventSource::ConnectionListener;
-
         Self {
             wait_handles: [NULL; MAX_EVENT_COUNT],
             sources: [DEFAULT_EVENT_SOURCE; MAX_EVENT_COUNT],
@@ -1067,12 +1093,15 @@ impl EventListener {
 
     pub fn track(&mut self, event: &Event, source: EventSource) {
         let index = self.len as usize;
-        debug_assert!(index < self.wait_handles.len());
+        if index >= self.wait_handles.len() {
+            panic!("tried to track too many events");
+        }
         self.wait_handles[index] = event.handle();
         self.sources[index] = source;
         self.len += 1;
     }
 
+    #[inline(never)]
     pub fn wait_next(&mut self, timeout: Option<Duration>) -> Option<EventSource> {
         let len = self.len;
         self.len = 0;
@@ -1481,7 +1510,7 @@ fn run_client(args: Args, pipe_path: &[u16]) {
 
     if let Some(handle) = &console_output_handle {
         let size = get_console_size(handle);
-        let (_, bytes) = application.update(Some(size), &[Key::None], None, &[]);
+        let (_, bytes) = application.update(Some(size), &[Key::default()], None, &[]);
         if !connection.write(bytes) {
             return;
         }
@@ -1548,7 +1577,12 @@ fn run_client(args: Args, pipe_path: &[u16]) {
                     parse_console_events(console_events, &mut keys, &mut resize);
                 }
             }
-            1 => keys.push(Key::Ctrl('c')),
+            1 => keys.push(Key {
+                code: KeyCode::Char('c'),
+                shift: false,
+                control: true,
+                alt: false,
+            }),
             2 => match connection.read_async() {
                 Ok(bytes) => server_bytes = bytes,
                 Err(()) => break,
@@ -1600,50 +1634,62 @@ fn parse_console_events(
                 let unicode_char = unsafe { *event.uChar.UnicodeChar() };
                 let repeat_count = event.wRepeatCount as usize;
 
+                let mut shift = control_key_state & SHIFT_PRESSED != 0;
+                let control = control_key_state & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED) != 0;
+                let alt = control_key_state & (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED) != 0;
+
                 const CHAR_A: i32 = b'A' as _;
                 const CHAR_Z: i32 = b'Z' as _;
-                let key = match keycode {
-                    VK_BACK => Key::Backspace,
-                    VK_RETURN => Key::Enter,
-                    VK_LEFT => Key::Left,
-                    VK_RIGHT => Key::Right,
-                    VK_UP => Key::Up,
-                    VK_DOWN => Key::Down,
-                    VK_HOME => Key::Home,
-                    VK_END => Key::End,
-                    VK_PRIOR => Key::PageUp,
-                    VK_NEXT => Key::PageDown,
-                    VK_TAB => Key::Tab,
-                    VK_DELETE => Key::Delete,
-                    VK_F1..=VK_F24 => Key::F((keycode - VK_F1 + 1) as _),
-                    VK_ESCAPE => Key::Esc,
-                    VK_SPACE => {
-                        match std::char::decode_utf16(std::iter::once(unicode_char)).next() {
-                            Some(Ok(c)) => Key::Char(c),
-                            _ => continue,
-                        }
-                    }
-                    CHAR_A..=CHAR_Z => {
-                        const ALT_PRESSED_MASK: DWORD = LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED;
-                        const CTRL_PRESSED_MASK: DWORD = LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED;
-
-                        if control_key_state & ALT_PRESSED_MASK != 0 {
-                            let c = (keycode - CHAR_A) as u8 + b'a';
-                            Key::Alt(c.to_ascii_lowercase() as _)
-                        } else if control_key_state & CTRL_PRESSED_MASK != 0 {
-                            let c = (keycode - CHAR_A) as u8 + b'a';
-                            Key::Ctrl(c.to_ascii_lowercase() as _)
-                        } else {
-                            match std::char::decode_utf16(std::iter::once(unicode_char)).next() {
-                                Some(Ok(c)) => Key::Char(c),
-                                _ => continue,
-                            }
-                        }
-                    }
-                    _ => match std::char::decode_utf16(std::iter::once(unicode_char)).next() {
-                        Some(Ok(c)) if c.is_ascii_graphic() => Key::Char(c),
+                let code = match keycode {
+                    VK_BACK => KeyCode::Backspace,
+                    VK_RETURN => KeyCode::Char('\n'),
+                    VK_LEFT => KeyCode::Left,
+                    VK_RIGHT => KeyCode::Right,
+                    VK_UP => KeyCode::Up,
+                    VK_DOWN => KeyCode::Down,
+                    VK_HOME => KeyCode::Home,
+                    VK_END => KeyCode::End,
+                    VK_PRIOR => KeyCode::PageUp,
+                    VK_NEXT => KeyCode::PageDown,
+                    VK_TAB => KeyCode::Char('\t'),
+                    VK_DELETE => KeyCode::Delete,
+                    VK_F1..=VK_F24 => KeyCode::F((keycode - VK_F1 + 1) as _),
+                    VK_ESCAPE => KeyCode::Esc,
+                    VK_SPACE => match std::char::decode_utf16(std::iter::once(unicode_char)).next()
+                    {
+                        Some(Ok(c)) => KeyCode::Char(c),
                         _ => continue,
                     },
+                    CHAR_A..=CHAR_Z => {
+                        let mut c = if control || alt {
+                            ((keycode - CHAR_A) as u8 + b'a') as char
+                        } else {
+                            match std::char::decode_utf16(std::iter::once(unicode_char)).next() {
+                                Some(Ok(c)) => c,
+                                _ => continue,
+                            }
+                        };
+                        if shift {
+                            c = c.to_ascii_uppercase();
+                        } else {
+                            shift = c.is_ascii_uppercase();
+                        }
+                        KeyCode::Char(c)
+                    }
+                    _ => match std::char::decode_utf16(std::iter::once(unicode_char)).next() {
+                        Some(Ok(c)) if c.is_ascii_graphic() => {
+                            shift = false;
+                            KeyCode::Char(c)
+                        }
+                        _ => continue,
+                    },
+                };
+
+                let key = Key {
+                    code,
+                    shift,
+                    control,
+                    alt,
                 };
 
                 for _ in 0..repeat_count {
